@@ -55,8 +55,8 @@ const BITS_PER_BYTE: crate::Bytes = 8;
 pub enum Error {
     #[error("target SPIRV-{0}.{1} is not supported")]
     UnsupportedVersion(u8, u8),
-    #[error("one of the required capabilities {0:?} is missing")]
-    MissingCapabilities(Vec<Capability>),
+    #[error("using {0} requires at least one of the capabilities {1:?}, but none are available")]
+    MissingCapabilities(&'static str, Vec<Capability>),
     #[error("unimplemented {0}")]
     FeatureNotImplemented(&'static str),
     #[error("module is not validated properly: {0}")]
@@ -151,6 +151,67 @@ impl Function {
     }
 }
 
+/// Characteristics of a SPIR-V `OpTypeImage` type.
+///
+/// SPIR-V requires non-composite types to be unique, including images. Since we
+/// use `LocalType` for this deduplication, it's essential that `LocalImageType`
+/// be equal whenever the corresponding `OpTypeImage`s would be. To reduce the
+/// likelihood of mistakes, we use fields that correspond exactly to the
+/// operands of an `OpTypeImage` instruction, using the actual SPIR-V types
+/// where practical.
+#[derive(Debug, PartialEq, Hash, Eq, Copy, Clone)]
+struct LocalImageType {
+    sampled_type: crate::ScalarKind,
+    dim: spirv::Dim,
+    flags: ImageTypeFlags,
+    image_format: spirv::ImageFormat,
+}
+
+bitflags::bitflags! {
+    /// Flags corresponding to the boolean(-ish) parameters to OpTypeImage.
+    pub struct ImageTypeFlags: u8 {
+        const DEPTH = 0x1;
+        const ARRAYED = 0x2;
+        const MULTISAMPLED = 0x4;
+        const SAMPLED = 0x8;
+    }
+}
+
+impl LocalImageType {
+    /// Construct a `LocalImageType` from the fields of a `TypeInner::Image`.
+    fn from_inner(dim: crate::ImageDimension, arrayed: bool, class: crate::ImageClass) -> Self {
+        let make_flags = |multi: bool, other: ImageTypeFlags| -> ImageTypeFlags {
+            let mut flags = other;
+            flags.set(ImageTypeFlags::ARRAYED, arrayed);
+            flags.set(ImageTypeFlags::MULTISAMPLED, multi);
+            flags
+        };
+
+        let dim = spirv::Dim::from(dim);
+
+        match class {
+            crate::ImageClass::Sampled { kind, multi } => LocalImageType {
+                sampled_type: kind,
+                dim,
+                flags: make_flags(multi, ImageTypeFlags::SAMPLED),
+                image_format: spirv::ImageFormat::Unknown,
+            },
+            crate::ImageClass::Depth { multi } => LocalImageType {
+                sampled_type: crate::ScalarKind::Float,
+                dim,
+                flags: make_flags(multi, ImageTypeFlags::DEPTH | ImageTypeFlags::SAMPLED),
+                image_format: spirv::ImageFormat::Unknown,
+            },
+            crate::ImageClass::Storage { format, access: _ } => LocalImageType {
+                sampled_type: crate::ScalarKind::from(format),
+                dim,
+                flags: make_flags(false, ImageTypeFlags::empty()),
+                image_format: format.into(),
+            },
+        }
+    }
+}
+
 /// A SPIR-V type constructed during code generation.
 ///
 /// In the process of writing SPIR-V, we need to synthesize various types for
@@ -159,6 +220,13 @@ impl Function {
 /// so we can't ever create a `Handle<Type>` to refer to them. So for local use
 /// in the SPIR-V writer, we have this home-grown type enum that covers only the
 /// cases we need (for example, it doesn't cover structs).
+///
+/// As explained in §2.8 of the SPIR-V spec, some classes of type instructions
+/// must be unique; for example, you can't have two `OpTypeInt 32 1`
+/// instructions in the same module. `Writer::lookup_type` maps each `LocalType`
+/// value for which we've written instructions to its id, so we can avoid
+/// writing out duplicates. `LocalType` also includes variants like `Pointer`
+/// that do not need to be unique - but it is harmless to avoid the duplication.
 #[derive(Debug, PartialEq, Hash, Eq, Copy, Clone)]
 enum LocalType {
     /// A scalar, vector, or pointer to one of those.
@@ -180,11 +248,7 @@ enum LocalType {
         base: Handle<crate::Type>,
         class: spirv::StorageClass,
     },
-    Image {
-        dim: crate::ImageDimension,
-        arrayed: bool,
-        class: crate::ImageClass,
-    },
+    Image(LocalImageType),
     SampledImage {
         image_type_id: Word,
     },
@@ -253,11 +317,7 @@ fn make_local(inner: &crate::TypeInner) -> Option<LocalType> {
             dim,
             arrayed,
             class,
-        } => LocalType::Image {
-            dim,
-            arrayed,
-            class,
-        },
+        } => LocalType::Image(LocalImageType::from_inner(dim, arrayed, class)),
         crate::TypeInner::Sampler { comparison: _ } => LocalType::Sampler,
         _ => return None,
     })
@@ -385,8 +445,18 @@ pub struct Writer {
     physical_layout: PhysicalLayout,
     logical_layout: LogicalLayout,
     id_gen: IdGenerator,
-    capabilities: crate::FastHashSet<Capability>,
-    forbidden_caps: Option<&'static [Capability]>,
+
+    /// The set of capabilities modules are permitted to use.
+    ///
+    /// This is initialized from `Options::capabilities`.
+    capabilities_available: Option<crate::FastHashSet<Capability>>,
+
+    /// The set of capabilities used by this module.
+    ///
+    /// If `capabilities_available` is `Some`, then this is always a subset of
+    /// that.
+    capabilities_used: crate::FastHashSet<Capability>,
+
     debugs: Vec<Instruction>,
     annotations: Vec<Instruction>,
     flags: WriterFlags,
@@ -422,12 +492,16 @@ bitflags::bitflags! {
 pub struct Options {
     /// (Major, Minor) target version of the SPIR-V.
     pub lang_version: (u8, u8),
+
     /// Configuration flags for the writer.
     pub flags: WriterFlags,
-    /// Set of SPIR-V allowed capabilities, if provided.
-    // Note: there is a major bug currently associated with deriving the capabilities.
-    // We are calling `required_capabilities`, but the semantics of this is broken.
+
+    /// If given, the set of capabilities modules are allowed to use. Code that
+    /// requires capabilities beyond these is rejected with an error.
+    ///
+    /// If this is `None`, all capabilities are permitted.
     pub capabilities: Option<crate::FastHashSet<Capability>>,
+
     /// How should the generated code handle array, vector, or matrix indices
     /// that are out of range?
     pub index_bounds_check_policy: IndexBoundsCheckPolicy,
