@@ -13,15 +13,6 @@ use crate::{
 use spirv::Word;
 use std::collections::hash_map::Entry;
 
-fn map_dim(dim: crate::ImageDimension) -> spirv::Dim {
-    match dim {
-        crate::ImageDimension::D1 => spirv::Dim::Dim1D,
-        crate::ImageDimension::D2 => spirv::Dim::Dim2D,
-        crate::ImageDimension::D3 => spirv::Dim::Dim3D,
-        crate::ImageDimension::Cube => spirv::Dim::DimCube,
-    }
-}
-
 impl Function {
     fn to_words(&self, sink: &mut impl Extend<Word>) {
         self.signature.as_ref().unwrap().to_words(sink);
@@ -50,15 +41,8 @@ impl Writer {
         }
         let raw_version = ((major as u32) << 16) | ((minor as u32) << 8);
 
-        let (capabilities, forbidden_caps) = match options.capabilities {
-            Some(ref caps) => (caps.clone(), None),
-            None => {
-                let mut caps = crate::FastHashSet::default();
-                caps.insert(spirv::Capability::Shader);
-                let forbidden: &[_] = &[spirv::Capability::Kernel];
-                (caps, Some(forbidden))
-            }
-        };
+        let mut capabilities_used = crate::FastHashSet::default();
+        capabilities_used.insert(spirv::Capability::Shader);
 
         let mut id_gen = IdGenerator::default();
         let gl450_ext_inst_id = id_gen.next();
@@ -68,8 +52,8 @@ impl Writer {
             physical_layout: PhysicalLayout::new(raw_version),
             logical_layout: LogicalLayout::default(),
             id_gen,
-            capabilities,
-            forbidden_caps,
+            capabilities_available: options.capabilities.clone(),
+            capabilities_used,
             debugs: vec![],
             annotations: vec![],
             flags: options.flags,
@@ -110,8 +94,7 @@ impl Writer {
             // Copied from the old Writer:
             flags: self.flags,
             index_bounds_check_policy: self.index_bounds_check_policy,
-            capabilities: take(&mut self.capabilities),
-            forbidden_caps: take(&mut self.forbidden_caps),
+            capabilities_available: take(&mut self.capabilities_available),
 
             // Initialized afresh:
             id_gen,
@@ -119,6 +102,7 @@ impl Writer {
             gl450_ext_inst_id,
 
             // Recycled:
+            capabilities_used: take(&mut self.capabilities_used).recycle(),
             physical_layout: self.physical_layout.clone().recycle(),
             logical_layout: take(&mut self.logical_layout).recycle(),
             debugs: take(&mut self.debugs).recycle(),
@@ -134,36 +118,64 @@ impl Writer {
         };
 
         *self = fresh;
+
+        self.capabilities_used.insert(spirv::Capability::Shader);
     }
 
-    //TODO: revive and rewrite this, see:
-    // https://github.com/gfx-rs/rspirv/issues/198
-    // https://github.com/gfx-rs/rspirv/pull/185#issuecomment-900796025
-    pub(super) fn check(&mut self, capabilities: &[spirv::Capability]) -> Result<(), Error> {
-        if capabilities.is_empty()
-            || capabilities
-                .iter()
-                .any(|cap| self.capabilities.contains(cap))
-        {
-            return Ok(());
-        }
-        if let Some(forbidden) = self.forbidden_caps {
-            // take the first allowed capability, blindly
-            if let Some(&cap) = capabilities.iter().find(|cap| !forbidden.contains(cap)) {
-                self.capabilities.insert(cap);
-                return Ok(());
+    /// Indicate that the code requires any one of the listed capabilities.
+    ///
+    /// If nothing in `capabilities` appears in the available capabilities
+    /// specified in the [`Options`] from which this `Writer` was created,
+    /// return an error. The `what` string is used in the error message to
+    /// explain what provoked the requirement. (If no available capabilites were
+    /// given, assume everything is available.)
+    ///
+    /// The first acceptable capability will be added to this `Writer`'s
+    /// [`capabilities_used`] table, and an `OpCapability` emitted for it in the
+    /// result. For this reason, more specific capabilities should be listed
+    /// before more general.
+    ///
+    /// [`capabilities_used`]: Writer::capabilities_used
+    pub(super) fn require_any(
+        &mut self,
+        what: &'static str,
+        capabilities: &[spirv::Capability],
+    ) -> Result<(), Error> {
+        match *capabilities {
+            [] => Ok(()),
+            [first, ..] => {
+                // Find the first acceptable capability, or return an error if
+                // there is none.
+                let selected = match self.capabilities_available {
+                    None => first,
+                    Some(ref available) => {
+                        match capabilities.iter().find(|cap| available.contains(cap)) {
+                            Some(&cap) => cap,
+                            None => {
+                                return Err(Error::MissingCapabilities(what, capabilities.to_vec()))
+                            }
+                        }
+                    }
+                };
+                self.capabilities_used.insert(selected);
+                Ok(())
             }
         }
-        Err(Error::MissingCapabilities(capabilities.to_vec()))
     }
 
     pub(super) fn get_type_id(&mut self, lookup_ty: LookupType) -> Word {
-        if let Entry::Occupied(e) = self.lookup_type.entry(lookup_ty) {
-            *e.get()
-        } else {
-            match lookup_ty {
-                LookupType::Handle(_handle) => unreachable!("Handles are populated at start"),
-                LookupType::Local(local_ty) => self.write_type_declaration_local(local_ty),
+        match self.lookup_type.entry(lookup_ty) {
+            Entry::Occupied(e) => *e.get(),
+            Entry::Vacant(e) => {
+                let local = match lookup_ty {
+                    LookupType::Handle(_handle) => unreachable!("Handles are populated at start"),
+                    LookupType::Local(local) => local,
+                };
+
+                let id = self.id_gen.next();
+                e.insert(id);
+                self.write_type_declaration_local(id, local);
+                id
             }
         }
     }
@@ -556,13 +568,13 @@ impl Writer {
                     _ => None,
                 };
                 if let Some(cap) = cap {
-                    self.capabilities.insert(cap);
+                    self.capabilities_used.insert(cap);
                 }
                 Instruction::type_int(id, bits, signedness)
             }
             Sk::Float => {
                 if bits == 64 {
-                    self.capabilities.insert(spirv::Capability::Float64);
+                    self.capabilities_used.insert(spirv::Capability::Float64);
                 }
                 Instruction::type_float(id, bits)
             }
@@ -570,8 +582,51 @@ impl Writer {
         }
     }
 
-    fn write_type_declaration_local(&mut self, local_ty: LocalType) -> Word {
-        let id = self.id_gen.next();
+    fn request_image_capabilities(&mut self, inner: &crate::TypeInner) -> Result<(), Error> {
+        if let crate::TypeInner::Image {
+            dim,
+            arrayed,
+            class,
+        } = *inner
+        {
+            let sampled = match class {
+                crate::ImageClass::Sampled { .. } => true,
+                crate::ImageClass::Depth { .. } => true,
+                crate::ImageClass::Storage { format, .. } => {
+                    self.request_image_format_capabilities(format.into())?;
+                    false
+                }
+            };
+
+            match dim {
+                crate::ImageDimension::D1 => {
+                    if sampled {
+                        self.require_any("sampled 1D images", &[spirv::Capability::Sampled1D])?;
+                    } else {
+                        self.require_any("1D storage images", &[spirv::Capability::Image1D])?;
+                    }
+                }
+                crate::ImageDimension::Cube if arrayed => {
+                    if sampled {
+                        self.require_any(
+                            "sampled cube array images",
+                            &[spirv::Capability::SampledCubeArray],
+                        )?;
+                    } else {
+                        self.require_any(
+                            "cube array storage images",
+                            &[spirv::Capability::ImageCubeArray],
+                        )?;
+                    }
+                }
+                _ => {}
+            }
+        }
+
+        Ok(())
+    }
+
+    fn write_type_declaration_local(&mut self, id: Word, local_ty: LocalType) {
         let instruction = match local_ty {
             LocalType::Value {
                 vector_size: None,
@@ -624,16 +679,23 @@ impl Writer {
                 }));
                 Instruction::type_pointer(id, class, type_id)
             }
-            // all the samplers and image types go through `write_type_declaration_arena`
-            LocalType::Image { .. } | LocalType::Sampler => unreachable!(),
+            LocalType::Image(image) => {
+                let local_type = LocalType::Value {
+                    vector_size: None,
+                    kind: image.sampled_type,
+                    width: 4,
+                    pointer_class: None,
+                };
+                let type_id = self.get_type_id(LookupType::Local(local_type));
+                Instruction::type_image(id, type_id, image.dim, image.flags, image.image_format)
+            }
+            LocalType::Sampler => Instruction::type_sampler(id),
             LocalType::SampledImage { image_type_id } => {
                 Instruction::type_sampled_image(id, image_type_id)
             }
         };
 
-        self.lookup_type.insert(LookupType::Local(local_ty), id);
         instruction.to_words(&mut self.logical_layout.declarations);
-        id
     }
 
     fn write_type_declaration_arena(
@@ -645,22 +707,125 @@ impl Writer {
         let decorate_layout = true; //TODO?
 
         let id = if let Some(local) = make_local(&ty.inner) {
+            // This type can be represented as a `LocalType`, so check if we've
+            // already written an instruction for it. If not, do so now, with
+            // `write_type_declaration_local`.
             match self.lookup_type.entry(LookupType::Local(local)) {
-                // if it's already known as local, re-use it
-                Entry::Occupied(e) => {
-                    let id = *e.into_mut();
-                    self.lookup_type.insert(LookupType::Handle(handle), id);
-                    return Ok(id);
-                }
-                // also register the type as "local", to avoid duplication
+                // We already have an id for this `LocalType`.
+                Entry::Occupied(e) => *e.get(),
+
+                // It's a type we haven't seen before.
                 Entry::Vacant(e) => {
                     let id = self.id_gen.next();
-                    *e.insert(id)
+                    e.insert(id);
+                    self.write_type_declaration_local(id, local);
+
+                    // If it's an image type, request SPIR-V capabilities here, so
+                    // write_type_declaration_local can stay infallible.
+                    self.request_image_capabilities(&ty.inner)?;
+
+                    id
                 }
             }
         } else {
-            self.id_gen.next()
+            use spirv::Decoration;
+
+            let id = self.id_gen.next();
+            let instruction = match ty.inner {
+                crate::TypeInner::Array { base, size, stride } => {
+                    if decorate_layout {
+                        self.decorate(id, Decoration::ArrayStride, &[stride]);
+                    }
+
+                    let type_id = self.get_type_id(LookupType::Handle(base));
+                    match size {
+                        crate::ArraySize::Constant(const_handle) => {
+                            let length_id = self.constant_ids[const_handle.index()];
+                            Instruction::type_array(id, type_id, length_id)
+                        }
+                        crate::ArraySize::Dynamic => Instruction::type_runtime_array(id, type_id),
+                    }
+                }
+                crate::TypeInner::Struct {
+                    top_level,
+                    ref members,
+                    span: _,
+                } => {
+                    if top_level {
+                        self.decorate(id, Decoration::Block, &[]);
+                    }
+
+                    let mut member_ids = Vec::with_capacity(members.len());
+                    for (index, member) in members.iter().enumerate() {
+                        if decorate_layout {
+                            self.annotations.push(Instruction::member_decorate(
+                                id,
+                                index as u32,
+                                Decoration::Offset,
+                                &[member.offset],
+                            ));
+                        }
+
+                        if self.flags.contains(WriterFlags::DEBUG) {
+                            if let Some(ref name) = member.name {
+                                self.debugs
+                                    .push(Instruction::member_name(id, index as u32, name));
+                            }
+                        }
+
+                        // The matrix decorations also go on arrays of matrices,
+                        // so lets check this first.
+                        let member_array_subty_inner = match arena[member.ty].inner {
+                            crate::TypeInner::Array { base, .. } => &arena[base].inner,
+                            ref other => other,
+                        };
+                        if let crate::TypeInner::Matrix {
+                            columns,
+                            rows: _,
+                            width,
+                        } = *member_array_subty_inner
+                        {
+                            let byte_stride = match columns {
+                                crate::VectorSize::Bi => 2 * width,
+                                crate::VectorSize::Tri | crate::VectorSize::Quad => 4 * width,
+                            };
+                            self.annotations.push(Instruction::member_decorate(
+                                id,
+                                index as u32,
+                                Decoration::ColMajor,
+                                &[],
+                            ));
+                            self.annotations.push(Instruction::member_decorate(
+                                id,
+                                index as u32,
+                                Decoration::MatrixStride,
+                                &[byte_stride as u32],
+                            ));
+                        }
+
+                        let member_id = self.get_type_id(LookupType::Handle(member.ty));
+                        member_ids.push(member_id);
+                    }
+                    Instruction::type_struct(id, member_ids.as_slice())
+                }
+
+                // These all have TypeLocal representations, so they should have been
+                // handled by `write_type_declaration_local` above.
+                crate::TypeInner::Scalar { .. }
+                | crate::TypeInner::Atomic { .. }
+                | crate::TypeInner::Vector { .. }
+                | crate::TypeInner::Matrix { .. }
+                | crate::TypeInner::Pointer { .. }
+                | crate::TypeInner::ValuePointer { .. }
+                | crate::TypeInner::Image { .. }
+                | crate::TypeInner::Sampler { .. } => unreachable!(),
+            };
+
+            instruction.to_words(&mut self.logical_layout.declarations);
+            id
         };
+
+        // Add this handle as a new alias for that type.
         self.lookup_type.insert(LookupType::Handle(handle), id);
 
         if self.flags.contains(WriterFlags::DEBUG) {
@@ -669,166 +834,65 @@ impl Writer {
             }
         }
 
-        use spirv::Decoration;
-
-        let instruction = match ty.inner {
-            crate::TypeInner::Scalar { kind, width } | crate::TypeInner::Atomic { kind, width } => {
-                self.make_scalar(id, kind, width)
-            }
-            crate::TypeInner::Vector { size, kind, width } => {
-                let scalar_id = self.get_type_id(LookupType::Local(LocalType::Value {
-                    vector_size: None,
-                    kind,
-                    width,
-                    pointer_class: None,
-                }));
-                Instruction::type_vector(id, scalar_id, size)
-            }
-            crate::TypeInner::Matrix {
-                columns,
-                rows,
-                width,
-            } => {
-                let vector_id = self.get_type_id(LookupType::Local(LocalType::Value {
-                    vector_size: Some(rows),
-                    kind: crate::ScalarKind::Float,
-                    width,
-                    pointer_class: None,
-                }));
-                Instruction::type_matrix(id, vector_id, columns)
-            }
-            crate::TypeInner::Image {
-                dim,
-                arrayed,
-                class,
-            } => {
-                let kind = match class {
-                    crate::ImageClass::Sampled { kind, multi: _ } => kind,
-                    crate::ImageClass::Depth { multi: _ } => crate::ScalarKind::Float,
-                    crate::ImageClass::Storage { format, .. } => {
-                        let required_caps: &[_] = match dim {
-                            crate::ImageDimension::D1 => &[spirv::Capability::Image1D],
-                            crate::ImageDimension::Cube => &[spirv::Capability::ImageCubeArray],
-                            _ => &[],
-                        };
-                        self.check(required_caps)?;
-                        format.into()
-                    }
-                };
-                let local_type = LocalType::Value {
-                    vector_size: None,
-                    kind,
-                    width: 4,
-                    pointer_class: None,
-                };
-                let dim = map_dim(dim);
-                //self.check(dim.required_capabilities())?;
-                let type_id = self.get_type_id(LookupType::Local(local_type));
-                Instruction::type_image(id, type_id, dim, arrayed, class)
-            }
-            crate::TypeInner::Sampler { comparison: _ } => Instruction::type_sampler(id),
-            crate::TypeInner::Array { base, size, stride } => {
-                if decorate_layout {
-                    self.decorate(id, Decoration::ArrayStride, &[stride]);
-                }
-
-                let type_id = self.get_type_id(LookupType::Handle(base));
-                match size {
-                    crate::ArraySize::Constant(const_handle) => {
-                        let length_id = self.constant_ids[const_handle.index()];
-                        Instruction::type_array(id, type_id, length_id)
-                    }
-                    crate::ArraySize::Dynamic => Instruction::type_runtime_array(id, type_id),
-                }
-            }
-            crate::TypeInner::Struct {
-                top_level,
-                ref members,
-                span: _,
-            } => {
-                if top_level {
-                    self.decorate(id, Decoration::Block, &[]);
-                }
-
-                let mut member_ids = Vec::with_capacity(members.len());
-                for (index, member) in members.iter().enumerate() {
-                    if decorate_layout {
-                        self.annotations.push(Instruction::member_decorate(
-                            id,
-                            index as u32,
-                            Decoration::Offset,
-                            &[member.offset],
-                        ));
-                    }
-
-                    if self.flags.contains(WriterFlags::DEBUG) {
-                        if let Some(ref name) = member.name {
-                            self.debugs
-                                .push(Instruction::member_name(id, index as u32, name));
-                        }
-                    }
-
-                    // The matrix decorations also go on arrays of matrices,
-                    // so lets check this first.
-                    let member_array_subty_inner = match arena[member.ty].inner {
-                        crate::TypeInner::Array { base, .. } => &arena[base].inner,
-                        ref other => other,
-                    };
-                    if let crate::TypeInner::Matrix {
-                        columns,
-                        rows: _,
-                        width,
-                    } = *member_array_subty_inner
-                    {
-                        let byte_stride = match columns {
-                            crate::VectorSize::Bi => 2 * width,
-                            crate::VectorSize::Tri | crate::VectorSize::Quad => 4 * width,
-                        };
-                        self.annotations.push(Instruction::member_decorate(
-                            id,
-                            index as u32,
-                            Decoration::ColMajor,
-                            &[],
-                        ));
-                        self.annotations.push(Instruction::member_decorate(
-                            id,
-                            index as u32,
-                            Decoration::MatrixStride,
-                            &[byte_stride as u32],
-                        ));
-                    }
-
-                    let member_id = self.get_type_id(LookupType::Handle(member.ty));
-                    member_ids.push(member_id);
-                }
-                Instruction::type_struct(id, member_ids.as_slice())
-            }
-            crate::TypeInner::Pointer { base, class } => {
-                let type_id = self.get_type_id(LookupType::Handle(base));
-                let raw_class = map_storage_class(class);
-                Instruction::type_pointer(id, raw_class, type_id)
-            }
-            crate::TypeInner::ValuePointer {
-                size,
-                kind,
-                width,
-                class,
-            } => {
-                let raw_class = map_storage_class(class);
-                let type_id = self.get_type_id(LookupType::Local(LocalType::Value {
-                    vector_size: size,
-                    kind,
-                    width,
-                    pointer_class: None,
-                }));
-                Instruction::type_pointer(id, raw_class, type_id)
-            }
-        };
-
-        instruction.to_words(&mut self.logical_layout.declarations);
         Ok(id)
     }
 
+    fn request_image_format_capabilities(
+        &mut self,
+        format: spirv::ImageFormat,
+    ) -> Result<(), Error> {
+        use spirv::ImageFormat as If;
+        match format {
+            If::Rg32f
+            | If::Rg16f
+            | If::R11fG11fB10f
+            | If::R16f
+            | If::Rgba16
+            | If::Rgb10A2
+            | If::Rg16
+            | If::Rg8
+            | If::R16
+            | If::R8
+            | If::Rgba16Snorm
+            | If::Rg16Snorm
+            | If::Rg8Snorm
+            | If::R16Snorm
+            | If::R8Snorm
+            | If::Rg32i
+            | If::Rg16i
+            | If::Rg8i
+            | If::R16i
+            | If::R8i
+            | If::Rgb10a2ui
+            | If::Rg32ui
+            | If::Rg16ui
+            | If::Rg8ui
+            | If::R16ui
+            | If::R8ui => self.require_any(
+                "storage image format",
+                &[spirv::Capability::StorageImageExtendedFormats],
+            ),
+            If::R64ui | If::R64i => self.require_any(
+                "64-bit integer storage image format",
+                &[spirv::Capability::Int64ImageEXT],
+            ),
+            If::Unknown
+            | If::Rgba32f
+            | If::Rgba16f
+            | If::R32f
+            | If::Rgba8
+            | If::Rgba8Snorm
+            | If::Rgba32i
+            | If::Rgba16i
+            | If::Rgba8i
+            | If::R32i
+            | If::Rgba32ui
+            | If::Rgba16ui
+            | If::Rgba8ui
+            | If::R32ui => Ok(()),
+        }
+    }
+
     pub(super) fn get_index_constant(&mut self, index: Word) -> Result<Word, Error> {
         self.get_constant_scalar(crate::ScalarValue::Uint(index as _), 4)
     }
@@ -1009,6 +1073,10 @@ impl Writer {
                         self.decorate(id, Decoration::Centroid, &[]);
                     }
                     Some(crate::Sampling::Sample) => {
+                        self.require_any(
+                            "per-sample interpolation",
+                            &[spirv::Capability::SampleRateShading],
+                        )?;
                         self.decorate(id, Decoration::Sample, &[]);
                     }
                 }
@@ -1035,10 +1103,20 @@ impl Writer {
                     Bi::FragDepth => BuiltIn::FragDepth,
                     Bi::FrontFacing => BuiltIn::FrontFacing,
                     Bi::PrimitiveIndex => {
-                        self.capabilities.insert(spirv::Capability::Geometry);
+                        self.require_any(
+                            "`primitive_index` built-in",
+                            &[spirv::Capability::Geometry],
+                        )?;
                         BuiltIn::PrimitiveId
                     }
-                    Bi::SampleIndex => BuiltIn::SampleId,
+                    Bi::SampleIndex => {
+                        self.require_any(
+                            "`sample_index` built-in",
+                            &[spirv::Capability::SampleRateShading],
+                        )?;
+
+                        BuiltIn::SampleId
+                    }
                     Bi::SampleMask => BuiltIn::SampleMask,
                     // compute
                     Bi::GlobalInvocationId => BuiltIn::GlobalInvocationId,
@@ -1222,7 +1300,7 @@ impl Writer {
             ep_instruction.to_words(&mut self.logical_layout.entry_points);
         }
 
-        for capability in self.capabilities.iter() {
+        for capability in self.capabilities_used.iter() {
             Instruction::capability(*capability).to_words(&mut self.logical_layout.capabilities);
         }
         if ir_module.entry_points.is_empty() {
@@ -1267,6 +1345,11 @@ impl Writer {
         self.logical_layout.in_words(words);
         Ok(())
     }
+
+    /// Return the set of capabilities the last module written used.
+    pub fn get_capabilities_used(&self) -> &crate::FastHashSet<spirv::Capability> {
+        &self.capabilities_used
+    }
 }
 
 #[test]
