@@ -1,12 +1,27 @@
+use super::{
+    ast::*,
+    context::Context,
+    error::{Error, ErrorKind},
+    Parser, Result, SourceMetadata,
+};
 use crate::{
-    Binding, Block, BuiltIn, Constant, Expression, GlobalVariable, Handle, ImageClass,
-    Interpolation, LocalVariable, ScalarKind, StorageAccess, StorageClass, SwizzleComponent, Type,
-    TypeInner, VectorSize,
+    Binding, Block, BuiltIn, Constant, Expression, GlobalVariable, Handle, Interpolation,
+    LocalVariable, ResourceBinding, ScalarKind, StorageAccess, StorageClass, SwizzleComponent,
+    Type, TypeInner, VectorSize,
 };
 
-use super::ast::*;
-use super::error::ErrorKind;
-use super::token::SourceMetadata;
+macro_rules! qualifier_arm {
+    ($src:expr, $tgt:expr, $meta:expr, $msg:literal, $errors:expr $(,)?) => {{
+        if $tgt.is_some() {
+            $errors.push(Error {
+                kind: ErrorKind::SemanticError($msg.into()),
+                meta: $meta,
+            })
+        }
+
+        $tgt = Some($src);
+    }};
+}
 
 pub struct VarDeclaration<'a> {
     pub qualifiers: &'a [(TypeQualifier, SourceMetadata)],
@@ -21,41 +36,44 @@ pub enum GlobalOrConstant {
     Constant(Handle<Constant>),
 }
 
-impl Program<'_> {
-    pub fn lookup_variable(
+impl Parser {
+    pub(crate) fn lookup_variable(
         &mut self,
         ctx: &mut Context,
         body: &mut Block,
         name: &str,
-    ) -> Result<Option<VariableReference>, ErrorKind> {
+        meta: SourceMetadata,
+    ) -> Option<VariableReference> {
         if let Some(local_var) = ctx.lookup_local_var(name) {
-            return Ok(Some(local_var));
+            return Some(local_var);
         }
         if let Some(global_var) = ctx.lookup_global_var(name) {
-            return Ok(Some(global_var));
+            return Some(global_var);
         }
 
-        let mut add_builtin = |inner, builtin, mutable, prologue| {
+        let mut add_builtin = |inner, builtin, mutable, storage| {
             let ty = self
                 .module
                 .types
-                .fetch_or_append(Type { name: None, inner });
+                .fetch_or_append(Type { name: None, inner }, meta.as_span());
 
-            let handle = self.module.global_variables.append(GlobalVariable {
-                name: Some(name.into()),
-                class: StorageClass::Private,
-                binding: None,
-                ty,
-                init: None,
-                storage_access: StorageAccess::empty(),
-            });
+            let handle = self.module.global_variables.append(
+                GlobalVariable {
+                    name: Some(name.into()),
+                    class: StorageClass::Private,
+                    binding: None,
+                    ty,
+                    init: None,
+                },
+                meta.as_span(),
+            );
 
             let idx = self.entry_args.len();
             self.entry_args.push(EntryArg {
                 name: None,
                 binding: Binding::BuiltIn(builtin),
                 handle,
-                prologue,
+                storage,
             });
 
             self.global_variables.push((
@@ -66,9 +84,8 @@ impl Program<'_> {
                     mutable,
                 },
             ));
-            ctx.arg_use.push(EntryArgUse::empty());
 
-            let expr = ctx.add_expression(Expression::GlobalVariable(handle), body);
+            let expr = ctx.add_expression(Expression::GlobalVariable(handle), meta, body);
             ctx.lookup_global_var_exps.insert(
                 name.into(),
                 VariableReference {
@@ -79,7 +96,7 @@ impl Program<'_> {
                 },
             );
 
-            Ok(ctx.lookup_global_var(name))
+            ctx.lookup_global_var(name)
         };
         match name {
             "gl_Position" => add_builtin(
@@ -90,25 +107,44 @@ impl Program<'_> {
                 },
                 BuiltIn::Position,
                 true,
-                PrologueStage::FRAGMENT,
+                StorageQualifier::Output,
+            ),
+            "gl_FragCoord" => add_builtin(
+                TypeInner::Vector {
+                    size: VectorSize::Quad,
+                    kind: ScalarKind::Float,
+                    width: 4,
+                },
+                BuiltIn::Position,
+                false,
+                StorageQualifier::Input,
+            ),
+            "gl_FragDepth" => add_builtin(
+                TypeInner::Scalar {
+                    kind: ScalarKind::Float,
+                    width: 4,
+                },
+                BuiltIn::FragDepth,
+                true,
+                StorageQualifier::Output,
             ),
             "gl_VertexIndex" => add_builtin(
                 TypeInner::Scalar {
-                    kind: ScalarKind::Sint,
+                    kind: ScalarKind::Uint,
                     width: 4,
                 },
                 BuiltIn::VertexIndex,
                 false,
-                PrologueStage::VERTEX,
+                StorageQualifier::Input,
             ),
             "gl_InstanceIndex" => add_builtin(
                 TypeInner::Scalar {
-                    kind: ScalarKind::Sint,
+                    kind: ScalarKind::Uint,
                     width: 4,
                 },
                 BuiltIn::InstanceIndex,
                 false,
-                PrologueStage::VERTEX,
+                StorageQualifier::Input,
             ),
             "gl_GlobalInvocationID" => add_builtin(
                 TypeInner::Vector {
@@ -118,22 +154,41 @@ impl Program<'_> {
                 },
                 BuiltIn::GlobalInvocationId,
                 false,
-                PrologueStage::COMPUTE,
+                StorageQualifier::Input,
+            ),
+            "gl_NumWorkGroups" => add_builtin(
+                TypeInner::Vector {
+                    size: VectorSize::Tri,
+                    kind: ScalarKind::Uint,
+                    width: 4,
+                },
+                BuiltIn::NumWorkGroups,
+                false,
+                StorageQualifier::Input,
             ),
             "gl_FrontFacing" => add_builtin(
                 TypeInner::Scalar {
                     kind: ScalarKind::Bool,
-                    width: 1,
+                    width: crate::BOOL_WIDTH,
                 },
                 BuiltIn::FrontFacing,
                 false,
-                PrologueStage::FRAGMENT,
+                StorageQualifier::Input,
             ),
-            _ => Ok(None),
+            "gl_PrimitiveID" => add_builtin(
+                TypeInner::Scalar {
+                    kind: ScalarKind::Uint,
+                    width: 4,
+                },
+                BuiltIn::PrimitiveIndex,
+                false,
+                StorageQualifier::Input,
+            ),
+            _ => None,
         }
     }
 
-    pub fn field_selection(
+    pub(crate) fn field_selection(
         &mut self,
         ctx: &mut Context,
         lhs: bool,
@@ -141,7 +196,7 @@ impl Program<'_> {
         expression: Handle<Expression>,
         name: &str,
         meta: SourceMetadata,
-    ) -> Result<Handle<Expression>, ErrorKind> {
+    ) -> Result<Handle<Expression>> {
         let (ty, is_pointer) = match *self.resolve_type(ctx, expression, meta)? {
             TypeInner::Pointer { base, .. } => (&self.module.types[base].inner, true),
             ref ty => (ty, false),
@@ -151,12 +206,16 @@ impl Program<'_> {
                 let index = members
                     .iter()
                     .position(|m| m.name == Some(name.into()))
-                    .ok_or_else(|| ErrorKind::UnknownField(meta, name.into()))?;
+                    .ok_or_else(|| Error {
+                        kind: ErrorKind::UnknownField(name.into()),
+                        meta,
+                    })?;
                 Ok(ctx.add_expression(
                     Expression::AccessIndex {
                         base: expression,
                         index: index as u32,
                     },
+                    meta,
                     body,
                 ))
             }
@@ -182,14 +241,17 @@ impl Program<'_> {
                         let not_unique = (1..components.len())
                             .any(|i| components[i..].contains(&components[i - 1]));
                         if not_unique {
-                            return Err(ErrorKind::SemanticError(
-                                meta,
+                            self.errors.push(Error {
+                                kind:
+                                ErrorKind::SemanticError(
                                 format!(
                                     "swizzle cannot have duplicate components in left-hand-side expression for \"{:?}\"",
                                     name
                                 )
                                 .into(),
-                            ));
+                            ),
+                                meta ,
+                            })
                         }
                     }
 
@@ -204,7 +266,7 @@ impl Program<'_> {
                         size: _,
                         vector,
                         pattern: ref src_pattern,
-                    } = *ctx.get_expression(expression)
+                    } = ctx[expression]
                     {
                         expression = vector;
                         for pat in &mut pattern {
@@ -217,9 +279,7 @@ impl Program<'_> {
                             // only single element swizzle, like pos.y, just return that component.
                             if lhs {
                                 // Because of possible nested swizzles, like pos.xy.x, we have to unwrap the potential load expr.
-                                if let Expression::Load { ref pointer } =
-                                    *ctx.get_expression(expression)
-                                {
+                                if let Expression::Load { ref pointer } = ctx[expression] {
                                     expression = *pointer;
                                 }
                             }
@@ -228,6 +288,7 @@ impl Program<'_> {
                                     base: expression,
                                     index: pattern[0].index(),
                                 },
+                                meta,
                                 body,
                             ));
                         }
@@ -235,10 +296,14 @@ impl Program<'_> {
                         3 => VectorSize::Tri,
                         4 => VectorSize::Quad,
                         _ => {
-                            return Err(ErrorKind::SemanticError(
+                            self.errors.push(Error {
+                                kind: ErrorKind::SemanticError(
+                                    format!("Bad swizzle size for \"{:?}\"", name).into(),
+                                ),
                                 meta,
-                                format!("Bad swizzle size for \"{:?}\"", name).into(),
-                            ));
+                            });
+
+                            VectorSize::Quad
                         }
                     };
 
@@ -250,6 +315,7 @@ impl Program<'_> {
                             Expression::Load {
                                 pointer: expression,
                             },
+                            meta,
                             body,
                         );
                     }
@@ -260,24 +326,31 @@ impl Program<'_> {
                             vector: expression,
                             pattern,
                         },
+                        meta,
                         body,
                     ))
                 } else {
-                    Err(ErrorKind::SemanticError(
+                    Err(Error {
+                        kind: ErrorKind::SemanticError(
+                            format!("Invalid swizzle for vector \"{}\"", name).into(),
+                        ),
                         meta,
-                        format!("Invalid swizzle for vector \"{}\"", name).into(),
-                    ))
+                    })
                 }
             }
-            _ => Err(ErrorKind::SemanticError(
+            _ => Err(Error {
+                kind: ErrorKind::SemanticError(
+                    format!("Can't lookup field on this type \"{}\"", name).into(),
+                ),
                 meta,
-                format!("Can't lookup field on this type \"{}\"", name).into(),
-            )),
+            }),
         }
     }
 
-    pub fn add_global_var(
+    pub(crate) fn add_global_var(
         &mut self,
+        ctx: &mut Context,
+        body: &mut Block,
         VarDeclaration {
             qualifiers,
             ty,
@@ -285,13 +358,16 @@ impl Program<'_> {
             init,
             meta,
         }: VarDeclaration,
-    ) -> Result<GlobalOrConstant, ErrorKind> {
+    ) -> Result<GlobalOrConstant> {
         let mut storage = StorageQualifier::StorageClass(StorageClass::Private);
         let mut interpolation = None;
+        let mut set = None;
         let mut binding = None;
         let mut location = None;
         let mut sampling = None;
         let mut layout = None;
+        let mut precision = None;
+        let mut access = StorageAccess::all();
 
         for &(ref qualifier, meta) in qualifiers {
             match *qualifier {
@@ -302,118 +378,139 @@ impl Program<'_> {
                         // Ignore the Uniform qualifier if the class was already set to PushConstant
                         continue;
                     } else if StorageQualifier::StorageClass(StorageClass::Private) != storage {
-                        return Err(ErrorKind::SemanticError(
+                        self.errors.push(Error {
+                            kind: ErrorKind::SemanticError(
+                                "Cannot use more than one storage qualifier per declaration".into(),
+                            ),
                             meta,
-                            "Cannot use more than one storage qualifier per declaration".into(),
-                        ));
+                        });
                     }
 
                     storage = s;
                 }
-                TypeQualifier::Interpolation(i) => {
-                    if interpolation.is_some() {
-                        return Err(ErrorKind::SemanticError(
-                            meta,
-                            "Cannot use more than one interpolation qualifier per declaration"
-                                .into(),
-                        ));
-                    }
-
-                    interpolation = Some(i);
-                }
-                TypeQualifier::ResourceBinding(ref r) => {
-                    if binding.is_some() {
-                        return Err(ErrorKind::SemanticError(
-                            meta,
-                            "Cannot use more than one binding per declaration".into(),
-                        ));
-                    }
-
-                    binding = Some(r.clone());
-                }
-                TypeQualifier::Location(l) => {
-                    if location.is_some() {
-                        return Err(ErrorKind::SemanticError(
-                            meta,
-                            "Cannot use more than one binding per declaration".into(),
-                        ));
-                    }
-
-                    location = Some(l);
-                }
-                TypeQualifier::Sampling(s) => {
-                    if sampling.is_some() {
-                        return Err(ErrorKind::SemanticError(
-                            meta,
-                            "Cannot use more than one sampling qualifier per declaration".into(),
-                        ));
-                    }
-
-                    sampling = Some(s);
-                }
-                TypeQualifier::Layout(ref l) => {
-                    if layout.is_some() {
-                        return Err(ErrorKind::SemanticError(
-                            meta,
-                            "Cannot use more than one layout qualifier per declaration".into(),
-                        ));
-                    }
-
-                    layout = Some(l);
-                }
+                TypeQualifier::Interpolation(i) => qualifier_arm!(
+                    i,
+                    interpolation,
+                    meta,
+                    "Cannot use more than one interpolation qualifier per declaration",
+                    self.errors
+                ),
+                TypeQualifier::Binding(r) => qualifier_arm!(
+                    r,
+                    binding,
+                    meta,
+                    "Cannot use more than one binding per declaration",
+                    self.errors
+                ),
+                TypeQualifier::Set(s) => qualifier_arm!(
+                    s,
+                    set,
+                    meta,
+                    "Cannot use more than one binding per declaration",
+                    self.errors
+                ),
+                TypeQualifier::Location(l) => qualifier_arm!(
+                    l,
+                    location,
+                    meta,
+                    "Cannot use more than one binding per declaration",
+                    self.errors
+                ),
+                TypeQualifier::Sampling(s) => qualifier_arm!(
+                    s,
+                    sampling,
+                    meta,
+                    "Cannot use more than one sampling qualifier per declaration",
+                    self.errors
+                ),
+                TypeQualifier::Layout(ref l) => qualifier_arm!(
+                    l,
+                    layout,
+                    meta,
+                    "Cannot use more than one layout qualifier per declaration",
+                    self.errors
+                ),
+                TypeQualifier::Precision(ref p) => qualifier_arm!(
+                    p,
+                    precision,
+                    meta,
+                    "Cannot use more than one precision qualifier per declaration",
+                    self.errors
+                ),
+                TypeQualifier::StorageAccess(a) => access &= a,
                 _ => {
-                    return Err(ErrorKind::SemanticError(
+                    self.errors.push(Error {
+                        kind: ErrorKind::SemanticError("Qualifier not supported in globals".into()),
                         meta,
-                        "Qualifier not supported in globals".into(),
-                    ));
+                    });
                 }
             }
         }
 
-        if binding.is_some() && storage != StorageQualifier::StorageClass(StorageClass::Uniform) {
-            match storage {
-                StorageQualifier::StorageClass(StorageClass::PushConstant)
-                | StorageQualifier::StorageClass(StorageClass::Uniform)
-                | StorageQualifier::StorageClass(StorageClass::Storage) => {}
-                _ => {
-                    return Err(ErrorKind::SemanticError(
+        match storage {
+            StorageQualifier::StorageClass(StorageClass::PushConstant) => {
+                if set.is_some() {
+                    self.errors.push(Error {
+                        kind: ErrorKind::SemanticError(
+                            "set cannot be used to decorate push constant".into(),
+                        ),
                         meta,
-                        "binding requires uniform or buffer storage qualifier".into(),
-                    ))
+                    })
+                }
+            }
+            StorageQualifier::StorageClass(StorageClass::Uniform)
+            | StorageQualifier::StorageClass(StorageClass::Storage { .. }) => {
+                if binding.is_none() {
+                    self.errors.push(Error {
+                        kind: ErrorKind::SemanticError(
+                            "uniform/buffer blocks require layout(binding=X)".into(),
+                        ),
+                        meta,
+                    })
+                }
+            }
+            _ => {
+                if set.is_some() || binding.is_some() {
+                    self.errors.push(Error {
+                        kind: ErrorKind::SemanticError(
+                            "set/binding can only be applied to uniform/buffer blocks".into(),
+                        ),
+                        meta,
+                    })
                 }
             }
         }
 
         if (sampling.is_some() || interpolation.is_some()) && location.is_none() {
-            return Err(ErrorKind::SemanticError(
+            return Err(Error {
+                kind: ErrorKind::SemanticError(
+                    "Sampling and interpolation qualifiers can only be used in in/out variables"
+                        .into(),
+                ),
                 meta,
-                "Sampling and interpolation qualifiers can only be used in in/out variables".into(),
-            ));
+            });
         }
 
         if let Some(location) = location {
             let input = storage == StorageQualifier::Input;
-            let prologue = if input {
-                PrologueStage::all()
-            } else {
-                PrologueStage::empty()
-            };
-            let interpolation = self.module.types[ty].inner.scalar_kind().map(|kind| {
-                if let ScalarKind::Float = kind {
-                    Interpolation::Perspective
-                } else {
-                    Interpolation::Flat
-                }
+            let interpolation = interpolation.or_else(|| {
+                let kind = self.module.types[ty].inner.scalar_kind()?;
+                Some(match kind {
+                    ScalarKind::Float => Interpolation::Perspective,
+                    _ => Interpolation::Flat,
+                })
             });
 
-            let handle = self.module.global_variables.append(GlobalVariable {
-                name: name.clone(),
-                class: StorageClass::Private,
-                binding: None,
-                ty,
-                init,
-                storage_access: StorageAccess::empty(),
-            });
+            let handle = self.module.global_variables.append(
+                GlobalVariable {
+                    name: name.clone(),
+                    class: StorageClass::Private,
+                    binding: None,
+                    ty,
+                    init,
+                },
+                meta.as_span(),
+            );
 
             let idx = self.entry_args.len();
             self.entry_args.push(EntryArg {
@@ -424,89 +521,83 @@ impl Program<'_> {
                     sampling,
                 },
                 handle,
-                prologue,
+                storage,
             });
 
             if let Some(name) = name {
-                self.global_variables.push((
-                    name,
-                    GlobalLookup {
-                        kind: GlobalLookupKind::Variable(handle),
-                        entry_arg: Some(idx),
-                        mutable: !input,
-                    },
-                ));
+                let lookup = GlobalLookup {
+                    kind: GlobalLookupKind::Variable(handle),
+                    entry_arg: Some(idx),
+                    mutable: !input,
+                };
+                ctx.add_global(self, &name, lookup, body);
+
+                self.global_variables.push((name, lookup));
             }
 
             return Ok(GlobalOrConstant::Global(handle));
         } else if let StorageQualifier::Const = storage {
-            let init = init.ok_or_else(|| {
-                ErrorKind::SemanticError(meta, "const values must have an initializer".into())
+            let init = init.ok_or_else(|| Error {
+                kind: ErrorKind::SemanticError("const values must have an initializer".into()),
+                meta,
             })?;
             if let Some(name) = name {
-                self.global_variables.push((
-                    name,
-                    GlobalLookup {
-                        kind: GlobalLookupKind::Constant(init),
-                        entry_arg: None,
-                        mutable: false,
-                    },
-                ));
+                let lookup = GlobalLookup {
+                    kind: GlobalLookupKind::Constant(init),
+                    entry_arg: None,
+                    mutable: false,
+                };
+                ctx.add_global(self, &name, lookup, body);
+
+                self.global_variables.push((name, lookup));
             }
             return Ok(GlobalOrConstant::Constant(init));
         }
 
-        let (class, storage_access) = match self.module.types[ty].inner {
-            TypeInner::Image { class, .. } => (
-                StorageClass::Handle,
-                if let ImageClass::Storage(_) = class {
-                    // TODO: Add support for qualifiers such as readonly,
-                    // writeonly and readwrite
-                    StorageAccess::all()
-                } else {
-                    StorageAccess::empty()
-                },
-            ),
-            TypeInner::Sampler { .. } => (StorageClass::Handle, StorageAccess::empty()),
+        let class = match self.module.types[ty].inner {
+            TypeInner::Image { .. } => StorageClass::Handle,
+            TypeInner::Sampler { .. } => StorageClass::Handle,
             _ => {
-                if let StorageQualifier::StorageClass(StorageClass::Storage) = storage {
-                    (StorageClass::Storage, StorageAccess::all())
+                if let StorageQualifier::StorageClass(StorageClass::Storage { .. }) = storage {
+                    StorageClass::Storage { access }
                 } else {
-                    (
-                        match storage {
-                            StorageQualifier::StorageClass(class) => class,
-                            _ => StorageClass::Private,
-                        },
-                        StorageAccess::empty(),
-                    )
+                    match storage {
+                        StorageQualifier::StorageClass(class) => class,
+                        _ => StorageClass::Private,
+                    }
                 }
             }
         };
 
-        let handle = self.module.global_variables.append(GlobalVariable {
-            name: name.clone(),
-            class,
-            binding,
-            ty,
-            init,
-            storage_access,
-        });
+        let handle = self.module.global_variables.append(
+            GlobalVariable {
+                name: name.clone(),
+                class,
+                binding: binding.map(|binding| ResourceBinding {
+                    group: set.unwrap_or(0),
+                    binding,
+                }),
+                ty,
+                init,
+            },
+            meta.as_span(),
+        );
 
         if let Some(name) = name {
-            self.global_variables.push((
-                name,
-                GlobalLookup {
-                    kind: GlobalLookupKind::Variable(handle),
-                    entry_arg: None,
-                    mutable: true,
-                },
-            ));
+            let lookup = GlobalLookup {
+                kind: GlobalLookupKind::Variable(handle),
+                entry_arg: None,
+                mutable: true,
+            };
+            ctx.add_global(self, &name, lookup, body);
+
+            self.global_variables.push((name, lookup));
         }
 
         Ok(GlobalOrConstant::Global(handle))
     }
 
-    pub fn add_local_var(
+    pub(crate) fn add_local_var(
         &mut self,
         ctx: &mut Context,
         body: &mut Block,
@@ -518,43 +609,58 @@ impl Program<'_> {
             init,
             meta,
         }: VarDeclaration,
-    ) -> Result<Handle<Expression>, ErrorKind> {
+    ) -> Result<Handle<Expression>> {
         #[cfg(feature = "glsl-validate")]
         if let Some(ref name) = name {
             if ctx.lookup_local_var_current_scope(name).is_some() {
-                return Err(ErrorKind::VariableAlreadyDeclared(meta, name.clone()));
+                self.errors.push(Error {
+                    kind: ErrorKind::VariableAlreadyDeclared(name.clone()),
+                    meta,
+                })
             }
         }
 
         let mut mutable = true;
+        let mut precision = None;
 
         for &(ref qualifier, meta) in qualifiers {
             match *qualifier {
                 TypeQualifier::StorageQualifier(StorageQualifier::Const) => {
                     if !mutable {
-                        return Err(ErrorKind::SemanticError(
+                        self.errors.push(Error {
+                            kind: ErrorKind::SemanticError(
+                                "Cannot use more than one constant qualifier per declaration"
+                                    .into(),
+                            ),
                             meta,
-                            "Cannot use more than one constant qualifier per declaration".into(),
-                        ));
+                        })
                     }
 
                     mutable = false;
                 }
-                _ => {
-                    return Err(ErrorKind::SemanticError(
-                        meta,
-                        "Qualifier not supported in locals".into(),
-                    ));
-                }
+                TypeQualifier::Precision(ref p) => qualifier_arm!(
+                    p,
+                    precision,
+                    meta,
+                    "Cannot use more than one precision qualifier per declaration",
+                    self.errors
+                ),
+                _ => self.errors.push(Error {
+                    kind: ErrorKind::SemanticError("Qualifier not supported in locals".into()),
+                    meta,
+                }),
             }
         }
 
-        let handle = ctx.locals.append(LocalVariable {
-            name: name.clone(),
-            ty,
-            init,
-        });
-        let expr = ctx.add_expression(Expression::LocalVariable(handle), body);
+        let handle = ctx.locals.append(
+            LocalVariable {
+                name: name.clone(),
+                ty,
+                init,
+            },
+            meta.as_span(),
+        );
+        let expr = ctx.add_expression(Expression::LocalVariable(handle), meta, body);
 
         if let Some(name) = name {
             ctx.add_local_var(name, expr, mutable);
